import * as Mqtt from 'mqtt'
import * as log from './log'
import * as callchain from './callchain'


const MQTT_BROKER_DEFAULT_URL : string = "mqtt://localhost"

export function passThrough(message: any) {
    return message
}



interface TopicHandler {
    topic: string,
    root: callchain.AChainItem | undefined
    last: callchain.AChainItem | undefined
}

export class MqttDispatcher {
    private _mqttClient: Mqtt.Client
    private _mqttBrokerUrl: string
    private _topicHandlers: TopicHandler[]

    constructor(mqttBrokerUrl? : string) {
        this._mqttBrokerUrl = (mqttBrokerUrl) ? mqttBrokerUrl : MQTT_BROKER_DEFAULT_URL
        this._topicHandlers = []
    }

    register(topic: string, label: string, 
            newChainItemOrCallbackFunc: callchain.AChainItem | callchain.ChainItemFunc) : void {
        let newChainItem : callchain.AChainItem
        if (newChainItemOrCallbackFunc instanceof callchain.AChainItem) {
            newChainItem = newChainItemOrCallbackFunc
        } else {
            let myNewChainItem = new callchain.ChainItem(label)
            myNewChainItem.registerFunc(newChainItemOrCallbackFunc)
            newChainItem = myNewChainItem
        }
        let done: boolean = false
        for (let topicHandler of this._topicHandlers) {
            if (topicHandler.topic === topic) {
                (topicHandler.last as callchain.AChainItem).registerNext(newChainItem)
                topicHandler.last = newChainItem
                done = true
                log.info(`additional callback ${newChainItem.toString()} added for topic ${topic}`)
            }
        }
        if (! done) {
            this._topicHandlers.push({topic: topic, root: newChainItem, last: newChainItem})
            log.info(`first callback ${newChainItem.toString()} added for topic ${topic}`)
        }
    }

    exec() : void {
        for (let topicHandler of this._topicHandlers) {
            (topicHandler.root as callchain.ChainItem).begin()
        }

        this._mqttClient = Mqtt.connect(this._mqttBrokerUrl)
        this._mqttClient.on('error', log.error)
        this._mqttClient.on('connect', (): void => {
            log.info("connected to mqtt broker")
            for (let topicHandler of this._topicHandlers) {
                this._mqttClient.subscribe(topicHandler.topic)
            }
        })
        this._mqttClient.on('message', (topic: string, payload: Buffer): void => {
            log.info(`message received, topic ${topic}, payload ${payload}`)
            for (let topicHandler of this._topicHandlers) {
                if (this.topicMatch(topicHandler.topic, topic)) {
                    log.info(`received topic ${topic} matches registered topic ${topicHandler.topic}`);
                    (topicHandler.root as callchain.ChainItem).send(payload)
                }
            }
        })
    }

    private topicMatch(registeredTopic: string, receivedTopic: string) : boolean {
        let registeredTopicFields = registeredTopic.split('/')
        let receivedTopicFields = receivedTopic.split('/')
        for (let field in registeredTopicFields) {
            let regField = registeredTopicFields[field]
            let recvField = receivedTopicFields[field]
            log.info(`recv: ${recvField}, reg: ${regField}`)
            if (regField === "#") {
                log.info('true')
                return true
            }
            if (regField != recvField && regField != "+") {
                log.info('false')
                return false
            }
        }
        if (registeredTopicFields.length == receivedTopicFields.length) {
            log.info('true')
            return true
        } else {
            log.info('false')
            return false
        }

    }
}
